import numpy as np
import torch
from bonito.util import stitch, concat

def stitch_results(results, length, size, overlap, stride, key, reverse=False):
    """
    Stitch results together with a given overlap.
    """
    if isinstance(results, dict):
        return {
            k: stitch_results(v, length, size, overlap, stride, k, reverse=reverse)
            for k, v in results.items()
        }
    if key == "segments":
        if length < size:
            return segments_single(results, size, overlap, length, stride, reverse=reverse)
   
        return stitch_segments_indices(results, size, overlap, length, stride, reverse=reverse)
    else:
        if length < size:
            return results[0, :int(np.floor(length / stride))]
        
        return stitch(results, size, overlap, length, stride, reverse=reverse)


def segments_single(chunks, chunksize, overlap, length, stride, reverse=False):
    n_chunks, l, n_peaks, per_peak = chunks.shape
    results = chunks[0, :int(np.floor(length / stride)),:,:]
    results = results[results != -2].view(-1, n_peaks, per_peak)
    return results


def stitch_segments_indices(chunks, chunksize, overlap, length, stride, reverse=False):
    """
    Stitch segments together with a given overlap
    """
    if chunks.shape[0] == 1: return chunks.squeeze(0)
    n_chunks, l, n_peaks, per_peak = chunks.shape

    semi_overlap = overlap // 2
    start, end = semi_overlap , (chunksize - semi_overlap) 
    stub = (length - overlap) % (chunksize - overlap)
    first_chunk_end = (stub + semi_overlap)  if (stub > 0) else end

    offset = torch.arange(0, (n_chunks-2)*(chunksize-overlap)+1, chunksize-overlap)
    offset += stub
    offset = torch.cat([torch.tensor([0]), offset]) # add offset for each chunk, since the segments are relative to the chunk start, not read start

    start_down = start // stride
    end_down = end // stride
    first_chunk_end_down = first_chunk_end // stride

    chunks[chunks == -2] = float("nan")
    chunks[:,:,:,0] += offset.view(-1,1,1)

    segments = concat([
            chunks[0, :first_chunk_end_down,:,:], *chunks[1:-1, start_down:end_down,:,:], chunks[-1, start_down:,:,:]
        ])
    segments = segments[~torch.isnan(segments)].view(-1, n_peaks, per_peak)

    segments = untangle_segments(segments.numpy())

    segments = make_end_dim(segments, length)

    return segments

def make_end_dim(segments, length):
    segments = np.stack([segments, np.roll(segments, shift=-1)], axis=-1)
    segments[-1,1] = length
    mask = np.any(segments == -1, axis=1)
    segments[mask,:] = -1
    return segments

def untangle_segments(segments):
    def best_fit(left, peaks, right):
        mask = (peaks[:, 0] > left) & (peaks[:, 0] < right)
        peaks = peaks[mask]
        if len(peaks) == 0:
            return -1,-1
        best_idx = np.argmax(peaks[:,1])
        return peaks[best_idx,:]
    
    def get_sort_shifts(arr):
        sorted_indices = np.argsort(arr)  # Get the indices that would sort the array
        original_indices = np.argsort(sorted_indices)  # Get the original positions after sorting
        shifts = original_indices - np.arange(len(arr))  # Compute shifts
        return np.abs(shifts)
    
    shifts = get_sort_shifts(segments[:,0,0])
    while sum(shifts) != 0:
        highest_shift = np.argmax(shifts)
        shifts[highest_shift] = 0


        left = np.max(segments[:highest_shift, 0, 0], initial=0) # max because there may be a 0 (no segment) in the one directly before/after, so just take the one that is not 0( since they are sorted its the max)
        right = np.max(segments[:highest_shift+2, 0, 0], initial=0) # is it correct to set the initial for right to 0? i think this is never needed?
        peaks = segments[highest_shift,:,:]

        segments[highest_shift, 0, :] = best_fit(left, peaks, right)
    # print(unsegmented)
    #TODO flag each "deleted" segment in the pandas dataframe, dont actually delete them, but remove them from the sorting, then once all of the remaining segments are sorted
    # go through the "deleted/ flagged" segments and see if any of the 5 peaks can be used to fit between the segment before and after, if it can, use it, if not, remove the
    # segment start property and say "it cant be positioned", see how many are problematic, if not many ok, if many do something else
    
    return segments[:,0,0]


import torch
from koi.decode import beam_search
from koi.ctc import Max

def run_beam_search(scores, beam_width=32, beam_cut=100.0, scale=1.0, offset=0.0, blank_score=2.0):
    with torch.cuda.device(scores.device):
        sequence, qstring, moves = beam_search( # IMPORTANT this is where the actual sequence is determined from the nn scores
            scores, beam_width=beam_width, beam_cut=beam_cut,
            scale=scale, offset=offset, blank_score=blank_score
        )
    return {
        'moves': moves,
        'qstring': qstring,
        'sequence': sequence,
    }


def number_to_dna_ascii(sequence):
    mapping = torch.tensor([0, 65, 67, 71, 84])  # [0, ord('A'), ord('C'), ord('G'), ord('T')]
    # Use index_select to map the values
    return mapping[sequence.to(torch.long)].to(torch.int8)

def viterbi(seqdist, scores):
    scores = seqdist.posteriors(scores.to(torch.float32)) + 1e-8 # probabilities
    traceback = seqdist.posteriors(scores.log(), Max) # one motif (last dimension) in traceback has score 1 and the rest 0
    a_traceback = traceback.argmax(2) #IMPORTANT take a_traceback (index of kmer (or base))
    return a_traceback

def run_viterbi(scores, seqdist, alphabet_len, n_base):
    # only use with use_koi = False
    scores_copy = scores.detach().clone()
    traceback = viterbi(seqdist, scores_copy)
    moves = (traceback % alphabet_len) != 0
    paths = 1 + (torch.div(traceback, alphabet_len, rounding_mode="floor") % n_base) # return which of the 4 bases
    paths = torch.where(moves, paths, 0).to("cpu").T # only get base when move TRUE -> move != 0
    paths_encoded = number_to_dna_ascii(paths)
    return {
        "sequence": paths_encoded,
        "moves": moves.to(torch.int8).to("cpu").T,
        "qstring": torch.zeros_like(paths),
        "traceback": traceback.to("cpu").T,
    }






"""
Bonito Basecaller
"""

import os
import sys
import numpy as np
from tqdm import tqdm
from time import perf_counter
from functools import partial
from datetime import timedelta
from itertools import islice as take
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from bonito.nn import fuse_bn_
from bonito.aligner import align_map, Aligner
from bonito.reader import read_chunks, Reader
from bonito.io import CTCWriter, Writer, biofmt
from bonito.mod_util import call_mods, load_mods_model
from bonito.cli.download import Downloader, models, __models_dir__
from bonito.multiprocessing import process_cancel, process_itemmap
from bonito.util import column_to_set, load_symbol, load_model, init, tqdm_environ

from relseg.lrp import basecall_and_lrp
from relseg.writer import LRP_Writer
from relseg.custom_model_loader import load_model_lrp
from relseg.transformer.nn import LinearCRFEncoder # just to register it


def main(args):
    search_algorithm = "viterbi"
    if search_algorithm == "viterbi":
        use_koi = False
    else:
        use_koi = True

    init(args.seed, args.device)

    try:
        reader = Reader(args.reads_directory, args.recursive)
        sys.stderr.write("> reading %s\n" % reader.fmt)
    except FileNotFoundError:
        sys.stderr.write("> error: no suitable files found in %s\n" % args.reads_directory)
        exit(1)

    fmt = biofmt(aligned=args.reference is not None)

    if args.reference and args.reference.endswith(".mmi") and fmt.name == "cram":
        sys.stderr.write("> error: reference cannot be a .mmi when outputting cram\n")
        exit(1)
    elif args.reference and fmt.name == "fastq":
        sys.stderr.write(f"> warning: did you really want {fmt.aligned} {fmt.name}?\n")
    else:
        sys.stderr.write(f"> outputting {fmt.aligned} {fmt.name}\n")

    if args.model_directory in models and not (__models_dir__ / args.model_directory).exists():
        sys.stderr.write("> downloading model\n")
        Downloader(__models_dir__).download(args.model_directory)

    if args.model_directory != "rna004_130bps_sup@v5.0.0":
        sys.stderr.write("> error: only model rna004_130bps_sup@v5.0.0 is tested\n")
        exit(1)

    sys.stderr.write(f"> loading model {args.model_directory}\n")
    try:
        model = load_model_lrp(
            args.model_directory,
            args.device,
            weights=args.weights if args.weights > 0 else None,
            chunksize=args.chunksize,
            overlap=args.overlap,
            batchsize=args.batchsize,
            quantize=args.quantize,
            use_koi=use_koi #TESTVALUE
        )
        model = model.apply(fuse_bn_)
    except FileNotFoundError:
        sys.stderr.write(f"> error: failed to load {args.model_directory}\n")
        sys.stderr.write(f"> available models:\n")
        for model in sorted(models): sys.stderr.write(f" - {model}\n")
        exit(1)

    if args.verbose:
        sys.stderr.write(f"> model basecaller params: {model.config['basecaller']}\n")


    mods_model = None
    if args.modified_base_model is not None or args.modified_bases is not None:
        sys.stderr.write("> loading modified base model\n")
        mods_model = load_mods_model(
            args.modified_bases, args.model_directory, args.modified_base_model,
            device=args.modified_device,
        )
        sys.stderr.write(f"> {mods_model[1]['alphabet_str']}\n")

    if args.reference:
        sys.stderr.write("> loading reference\n")
        aligner = Aligner(args.reference, preset=args.mm2_preset)
        if not aligner:
            sys.stderr.write("> failed to load/build index\n")
            exit(1)
    else:
        aligner = None

    if args.save_ctc and not args.reference:
        sys.stderr.write("> a reference is needed to output ctc training data\n")
        exit(1)

    if fmt.name != 'fastq':
        groups, num_reads = reader.get_read_groups(
            args.reads_directory, args.model_directory,
            n_proc=1, recursive=args.recursive,
            read_ids=column_to_set(args.read_ids), skip=args.skip,
            cancel=process_cancel()
        )
    else:
        groups = []
        num_reads = None

    reads = reader.get_reads(
        args.reads_directory, n_proc=1, recursive=args.recursive,
        read_ids=column_to_set(args.read_ids), skip=args.skip,
        do_trim=not args.no_trim,
        scaling_strategy=model.config.get("scaling"),
        norm_params=(model.config.get("standardisation")
                     if (model.config.get("scaling") and
                         model.config.get("scaling").get("strategy") == "pa")
                     else model.config.get("normalisation")
                     ),
        cancel=process_cancel()
    )

    if args.verbose:
        sys.stderr.write(f"> read scaling: {model.config.get('scaling')}\n")
    
    if args.max_reads:
        reads = take(reads, args.max_reads)
        if num_reads is not None:
            num_reads = min(num_reads, args.max_reads)

    if args.save_relevance:
        if not os.path.exists("relevance"): # not very elegant
            os.makedirs("relevance")


    results = basecall_and_lrp(
        model, reads, reverse=False, rna=args.rna,
        batchsize=model.config["basecaller"]["batchsize"],
        chunksize=model.config["basecaller"]["chunksize"],
        overlap=model.config["basecaller"]["overlap"],
        search_algorithm=search_algorithm,
        save_test_relevance=args.save_relevance
    )
    
    if mods_model is not None:
        if args.modified_device:
            results = ((k, call_mods(mods_model, k, v)) for k, v in results)
        else:
            results = process_itemmap(
                partial(call_mods, mods_model), results, n_proc=args.modified_procs
            )
    if aligner:
        results = align_map(aligner, results, n_thread=args.alignment_threads)

    writer_kwargs = {'aligner': aligner,
                     'group_key': args.model_directory,
                     'ref_fn': args.reference,
                     'groups': groups,
                     'min_qscore': args.min_qscore}
    if args.save_ctc:
        writer_kwargs['rna'] = args.rna
        writer_kwargs['min_accuracy'] = args.min_accuracy_save_ctc
        
    writer = LRP_Writer(
        fmt.mode, tqdm(results, desc="> calling", unit=" reads", leave=False,
                       total=num_reads, smoothing=0, ascii=True, ncols=100,
                       **tqdm_environ()),
        **writer_kwargs)

    t0 = perf_counter()
    #CHANGE
    # writer.start()
    # writer.join()
    writer.run()
    duration = perf_counter() - t0
    num_samples = sum(num_samples for read_id, num_samples in writer.log)

    sys.stderr.write("> completed reads: %s\n" % len(writer.log))
    sys.stderr.write("> duration: %s\n" % timedelta(seconds=np.round(duration)))
    sys.stderr.write("> samples per second %.1E\n" % (num_samples / duration))
    sys.stderr.write("> done\n")


def argparser():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument("model_directory")
    parser.add_argument("reads_directory")
    parser.add_argument("--reference")
    parser.add_argument("--modified-bases", nargs="+")
    parser.add_argument("--modified-base-model")
    parser.add_argument("--modified-procs", default=8, type=int)
    parser.add_argument("--modified-device", default=None)
    parser.add_argument("--read-ids")
    parser.add_argument("--device", default="cuda")
    parser.add_argument("--seed", default=25, type=int)
    parser.add_argument("--weights", default=0, type=int)
    parser.add_argument("--skip", action="store_true", default=False)
    parser.add_argument("--no-trim", action="store_true", default=False)
    parser.add_argument("--save-ctc", action="store_true", default=False)
    #parser.add_argument("--revcomp", action="store_true", default=False)
    parser.add_argument("--rna", action="store_true", default=False)
    parser.add_argument("--recursive", action="store_true", default=False)
    quant_parser = parser.add_mutually_exclusive_group(required=False)
    quant_parser.add_argument("--quantize", dest="quantize", action="store_true")
    quant_parser.add_argument("--no-quantize", dest="quantize", action="store_false")
    parser.set_defaults(quantize=None)
    parser.add_argument("--overlap", default=None, type=int)
    parser.add_argument("--chunksize", default=None, type=int)
    parser.add_argument("--batchsize", default=None, type=int)
    parser.add_argument("--max-reads", default=0, type=int)
    parser.add_argument("--min-qscore", default=0, type=int)
    parser.add_argument("--min-accuracy-save-ctc", default=0.99, type=float)
    parser.add_argument("--alignment-threads", default=8, type=int)
    parser.add_argument("--mm2-preset", default='lr:hq', type=str)
    parser.add_argument('-v', '--verbose', action='count', default=0)
    parser.add_argument('--save-relevance',action="store_true", default=False)
    return parser




import torch
import torch.nn as nn
import inspect
import lxt.functional as lf
from lxt.modules import INIT_MODULE_MAPPING, initialize_bias
import torch.fx
from torch.autograd import Function
from lxt.functional import conservation_check_wrap
import torch.nn.functional as F



class LinearAlphaBeta(nn.Linear):

    def __init__(self, in_features: int, out_features: int, bias: bool = True, device=None, dtype=None, alpha=1, **kwargs):
        super().__init__(in_features, out_features, bias, device, dtype)
        self.alpha = alpha

    def forward(self, inputs):
        return linear_alpha_beta(inputs, self.weight, self.bias, self.alpha)
    
@torch.fx.wrap
def linear_alpha_beta(input, weight, bias=None, alpha=0.8):
    """

    Parameters:
    -----------
    input: torch.Tensor
        The input tensor
    weight: torch.Tensor
        The weight tensor
    bias: torch.Tensor
        The bias tensor
    epsilon: float
        Small value to stabilize the denominator
    """
    return linear_alpha_beta_fn.apply(input, weight, bias, alpha)

#-------------------------------Inspiration from Transformer Interpretability Beyond Attention Visualization---------------
def safe_divide(a, b):
    den = b.clamp(min=1e-9) + b.clamp(max=1e-9)
    den = den + den.eq(0).type(den.type()) * 1e-9
    return a / den * b.ne(0).type(b.type())

class linear_alpha_beta_fn(Function):

    @staticmethod
    def forward(ctx, inputs, weight, bias=None, alpha=0.8):
        outputs = F.linear(inputs, weight, bias)
        ctx.save_for_backward(inputs, weight, outputs)
        ctx.alpha = alpha

        return outputs

    @staticmethod
    @conservation_check_wrap
    def backward(ctx, *out_relevance):

        inputs, weight, outputs = ctx.saved_tensors
        beta = ctx.alpha - 1
        pw = torch.clamp(weight, min=0)
        nw = torch.clamp(weight, max=0)
        px = torch.clamp(inputs, min=0)
        nx = torch.clamp(inputs, max=0)

        def f(w1, w2, x1, x2):
            Z1 = F.linear(x1, w1)
            Z2 = F.linear(x2, w2)
            S1 = safe_divide(out_relevance[0], Z1 + Z2)
            S2 = safe_divide(out_relevance[0], Z1 + Z2)
            C1 = x1 * (torch.matmul(S1, w1))
            C2 = x2 * (torch.matmul(S2, w2))
            # C1 above and C1 below are the same
            # C1 = x1 * torch.autograd.grad(Z1, x1, S1)[0]
            # C2 = x2 * torch.autograd.grad(Z2, x2, S2)[0]

            return C1 + C2
        
        activator_relevances = f(pw, nw, px, nx)
        inhibitor_relevances = f(nw, pw, px, nx)

        relevance = ctx.alpha * activator_relevances - beta * inhibitor_relevances


        # relevance_norm = out_relevance[0] / _stabilize(outputs, epsilon)

        # relevance = torch.matmul(relevance_norm, weight).mul_(inputs)

        return (relevance, None, None, None)

INIT_MODULE_MAPPING[LinearAlphaBeta] = initialize_bias


import torch
import torch.nn as nn
from lxt.functional import conservation_check_wrap


class RMSNorm(nn.Module):
    def __init__(self, d, p=-1., eps=torch.tensor(1e-8), bias=False):
        """
            Root Mean Square Layer Normalization
        :param d: model size
        :param p: partial RMSNorm, valid value [0, 1], default -1.0 (disabled)
        :param eps:  epsilon value, default 1e-8
        :param bias: whether use bias term for RMSNorm, disabled by
            default because RMSNorm doesn't enforce re-centering invariance.
        """
        super(RMSNorm, self).__init__()

        self.eps = eps
        self.d = d
        self.p = p
        self.bias = bias

        self.scale = nn.Parameter(torch.ones(d))
        self.register_parameter("scale", self.scale)

        # if self.bias:
        #     self.offset = nn.Parameter(torch.zeros(d))
        #     self.register_parameter("offset", self.offset)

    def forward(self, x=None):
        return rms_norm_identity_custom(x, self.scale, self.eps, self.d ,self.p, self.bias)
    
@torch.fx.wrap
def rms_norm_identity_custom(x, scale, epsilon, d, p, bias):
    return rms_norm_identity_function.apply(x, scale, epsilon, d, p, bias)


class rms_norm_identity_function(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, scale, epsilon, d, p, bias):

        if p < 0. or p > 1.:
            norm_x = x.norm(2, dim=-1, keepdim=True)
            d_x = d
        else:
            partial_size = int(d * p)
            partial_x, _ = torch.split(x, [partial_size, d - partial_size], dim=-1)

            norm_x = partial_x.norm(2, dim=-1, keepdim=True)
            d_x = partial_size

        rms_x = norm_x * d_x ** (-1. / 2)
        x_normed = x / (rms_x + epsilon)


        return scale * x_normed
    

    @staticmethod
    @conservation_check_wrap
    def backward(ctx, *out_relevance):

        #print(out_relevance)
        return out_relevance + (None, None, None, None, None, None)


# Copyright (c) 2023, Tri Dao.
# taken from flash_attn with minor changes

import torch

import relseg.lxt_adjustments as lf



def rotate_half(x, interleaved=False):
    if not interleaved:
        x1, x2 = x.chunk(2, dim=-1)
        return torch.cat((-x2, x1), dim=-1)
    else:
        x1, x2 = x[..., ::2], x[..., 1::2]
        return torch.stack((-x2, x1), dim=-1).reshape(*x1.shape[:-1], -1)

@torch.fx.wrap
def apply_rotary_emb_torch(x, cos, sin, interleaved=False):
    """
    x: (batch_size, seqlen, nheads, headdim)
    cos, sin: (seqlen, rotary_dim / 2) or (batch_size, seqlen, rotary_dim / 2)
    """
    ro_dim = cos.shape[-1] * 2
    #ssert ro_dim <= x.shape[-1]
    # cos = repeat(cos, "... d -> ... 1 (2 d)" if not interleaved else "... d -> ... 1 (d 2)")
    # sin = repeat(sin, "... d -> ... 1 (2 d)" if not interleaved else "... d -> ... 1 (d 2)")
    cos = cos.repeat(1,2)[:,None,:]
    sin = sin.repeat(1,2)[:,None,:]
    x_rotated = rotate_half(x[..., :ro_dim], interleaved)
    cos_x = lf.mul2(x, cos.detach())
    sin_x = lf.mul2(x_rotated, sin.detach())
    out = lf.add2(cos_x, sin_x)
    return out # IMPORTANT sin/cos show up in the table of functinos/modules of lxt but should not matter because they are detached?


class RotaryEmbedding(torch.nn.Module):
    """
    The rotary position embeddings from RoFormer_ (Su et. al).
    A crucial insight from the method is that the query and keys are
    transformed by rotation matrices which depend on the relative positions.

    Other implementations are available in the Rotary Transformer repo_ and in
    GPT-NeoX_, GPT-NeoX was an inspiration

    .. _RoFormer: https://arxiv.org/abs/2104.09864
    .. _repo: https://github.com/ZhuiyiTechnology/roformer
    .. _GPT-NeoX: https://github.com/EleutherAI/gpt-neox

    If scale_base is not None, this implements XPos (Sun et al., https://arxiv.org/abs/2212.10554).
    A recommended value for scale_base is 512: https://github.com/HazyResearch/flash-attention/issues/96
    Reference: https://github.com/sunyt32/torchscale/blob/main/torchscale/component/xpos_relative_position.py
    """

    def __init__(
        self,
        dim: int,
        base=10000.0,
        interleaved=False,
        scale_base=None,
        pos_idx_in_fp32=True,
        device=None,
    ):
        """
        interleaved: if True, rotate pairs of even and odd dimensions (GPT-J style) instead
            of 1st half and 2nd half (GPT-NeoX style).
        pos_idx_in_fp32: if True, the position indices [0.0, ..., seqlen - 1] are in fp32,
            otherwise they might be in lower precision.
            This option was added because previously (before 2023-07-02), when we construct
            the position indices, we use the dtype of self.inv_freq. In most cases this would
            be fp32, but if the model is trained in pure bf16 (not mixed precision), then
            self.inv_freq would be bf16, and the position indices are also in bf16.
            Because of the limited precision of bf16 (e.g. 1995.0 is rounded to 2000.0), the
            embeddings for some positions will coincide.
            To maintain compatibility with models previously trained in pure bf16,
            we add this option.
        """
        super().__init__()
        self.dim = dim
        self.base = float(base)
        self.pos_idx_in_fp32 = pos_idx_in_fp32
        # Generate and save the inverse frequency buffer (non trainable)
        inv_freq = self._compute_inv_freq(device)
        self.register_buffer("inv_freq", inv_freq, persistent=False)
        self.interleaved = interleaved
        self.scale_base = scale_base
        scale = (
            (torch.arange(0, dim, 2, device=device, dtype=torch.float32) + 0.4 * dim) / (1.4 * dim)
            if scale_base is not None
            else None
        )
        self.register_buffer("scale", scale, persistent=False)

        self._seq_len_cached = 0
        self._cos_cached = None
        self._sin_cached = None
        self._cos_k_cached = None
        self._sin_k_cached = None

    def _compute_inv_freq(self, device=None):
        return 1.0 / (
            self.base
            ** (torch.arange(0, self.dim, 2, device=device, dtype=torch.float32) / self.dim)
        )
    
    def apply_rotary_embedding_not_flash_qkv(self, qkv): # IMPORTANT this is a very unstable implementation (concerning shape of input, dtype, device...)
        self._update_cos_sin_cache(qkv.shape[1], device="cuda", dtype=qkv.dtype)
        def apply(x):
            return apply_rotary_emb_torch(x, self._cos_cached, self._sin_cached, self.interleaved)
        q_rotated = apply(qkv[:,:,0,:,:])
        k_rotated = apply(qkv[:,:,1,:,:])
        #v_rotated = apply_rotary_torch(qkv[:,:,2,:,:]) # v is not rotated
        qkv_rotated = torch.stack([q_rotated, k_rotated, qkv[:,:,2,:,:]], dim=2)
        return qkv_rotated
    
    def apply_rotary_embedding_not_flash_x(self, x):
        self._update_cos_sin_cache(x.shape[1], device="cuda", dtype=x.dtype)
        return apply_rotary_emb_torch(x, self._cos_cached, self._sin_cached, self.interleaved)

    def _update_cos_sin_cache(self, seqlen, device=None, dtype=None):
        # Reset the tables if the sequence length has changed,
        # if we're on a new device (possibly due to tracing for instance),
        # or if we're switching from inference mode to training

        self._seq_len_cached = seqlen
        # We want fp32 here, not self.inv_freq.dtype, since the model could be loaded in bf16
        # And the output of arange can be quite large, so bf16 would lose a lot of precision.
        # However, for compatibility reason, we add an option to use the dtype of self.inv_freq.
        if self.pos_idx_in_fp32:
            t = torch.arange(seqlen, device=device, dtype=torch.float32)
            # We want fp32 here as well since inv_freq will be multiplied with t, and the output
            # will be large. Having it in bf16 will lose a lot of precision and cause the
            # cos & sin output to change significantly.
            # We want to recompute self.inv_freq if it was not loaded in fp32
            if self.inv_freq.dtype != torch.float32:
                inv_freq = self._compute_inv_freq(device=device)
            else:
                inv_freq = self.inv_freq
        else:
            t = torch.arange(seqlen, device=device, dtype=self.inv_freq.dtype)
            inv_freq = self.inv_freq
        # Don't do einsum, it converts fp32 to fp16 under AMP
        # freqs = torch.einsum("i,j->ij", t, self.inv_freq)
        freqs = torch.outer(t, inv_freq)
        if self.scale is None:
            self._cos_cached = torch.cos(freqs).to(dtype)
            self._sin_cached = torch.sin(freqs).to(dtype)
            # self._cos_cached = freqs.to(dtype)
            # self._sin_cached = freqs.to(dtype)
        else:
            power = (
                torch.arange(seqlen, dtype=self.scale.dtype, device=self.scale.device)
                - seqlen // 2
            ) / self.scale_base
            scale = self.scale.to(device=power.device) ** power.unsqueeze(-1)
            # We want the multiplication by scale to happen in fp32
            self._cos_cached = (torch.cos(freqs) * scale).to(dtype)
            self._sin_cached = (torch.sin(freqs) * scale).to(dtype)
            self._cos_k_cached = (torch.cos(freqs) / scale).to(dtype)
            self._sin_k_cached = (torch.sin(freqs) / scale).to(dtype)




import operator
import torch
import torch.nn as nn


from lxt.core import Composite
import lxt.functional as lf
import lxt.modules as lm
import lxt.rules as rules
from .lrp_rules import LinearAlphaBeta


from zennit.composites import LayerMapComposite
import zennit.rules as z_rules

#####################
### LRP Composite ###
#####################

# class AttentionValueMatmul(nn.Module):
#     def forward(self, attn, value):
#         return torch.matmul(attn, value)

# class ProjSwigluMultiplication(nn.Module):
#     def forward(self, a, b):
#         return a * b

# must be traced because of the functions, e.g. operator.add!

conv_gamma = 0.01

zennit_comp1 = LayerMapComposite([
    (nn.Conv1d, z_rules.Gamma(conv_gamma))
])
zennit_comp2 = LayerMapComposite([
    (nn.Conv1d, z_rules.Gamma(conv_gamma))
])
zennit_comp3 = LayerMapComposite([
    (nn.Conv1d, z_rules.Gamma(conv_gamma))
])
zennit_comp4 = LayerMapComposite([
    (nn.Conv1d, z_rules.Gamma(conv_gamma))
])

zennit_comp_first_conv = LayerMapComposite([
    (nn.Conv1d, z_rules.WSquare()),
])

lxt_comp = Composite({

        nn.SiLU: rules.IdentityRule,
        nn.Tanh: rules.IdentityRule,
        nn.Sigmoid: rules.IdentityRule,

        # ProjSwigluMultiplication: rules.UniformRule,
        # AttentionValueMatmul: rules.UniformEpsilonRule,

        nn.Linear: LinearAlphaBeta,
        # nn.Conv1d: rules.UniformRule,

        nn.Softmax: lm.SoftmaxDT,
        # operator.add: lf.add2,
        # torch.add: lf.add2,
        # operator.mul: lf.mul2,
        # operator.truediv: lf.div2,
        # operator.matmul: lf.matmul,
        # torch.matmul: lf.matmul,
    })




import relseg.lxt_adjustments as lf
from bonito.nn import *

#this replaces the crfencoder, one multiplication has to be replaces
@register
class LinearCRFEncoder(Module):

    def __init__(self, insize, n_base, state_len, bias=True, scale=None, activation=None, blank_score=None, expand_blanks=True, permute=None):
        super().__init__()
        self.scale = scale
        self.n_base = n_base
        self.state_len = state_len
        self.blank_score = blank_score
        self.expand_blanks = expand_blanks
        size = (n_base + 1) * n_base**state_len if blank_score is None else n_base**(state_len + 1)
        self.linear = torch.nn.Linear(insize, size, bias=bias)
        self.activation = layers.get(activation, lambda: activation)()
        self.permute = permute

    def forward(self, x):
        if self.permute is not None:
            x = x.permute(*self.permute)
        scores = self.linear(x) #IMPORTANT this linear layer makes every single value negative. this must have something to do with the beamsearch afterwards and the training
        if self.activation is not None:
            scores = self.activation(scores)
        if self.scale is not None:
            scores = lf.mul2(scores, self.scale) # TESTVALUE
        if self.blank_score is not None and self.expand_blanks: # this adds the "empty/skip" base filled with blank scores: ACTG -> ACGTN
            T, N, C = scores.shape
            scores_view = scores.view(T, N, C // self.n_base, self.n_base)
            scores = torch.nn.functional.pad(
                scores_view,
                (1, 0, 0, 0, 0, 0, 0, 0),
                value=self.blank_score
            ) # -> (T, N, C // self.n_base, self.n_base+1) <-> adding the "N" dimension next to "ACGT" and giving it the value "blank_score"
            scores = scores.view(T, N, -1)
        return scores



from bonito.util import *
from bonito.util import __models_dir__
from bonito.util import _load_model

def load_model_lrp(dirname, device, weights=None, half=None, chunksize=None, batchsize=None, overlap=None, quantize=False, use_koi=False):
    """
    Load a model config and weights off disk from `dirname`.
    """
    if not os.path.isdir(dirname) and os.path.isdir(os.path.join(__models_dir__, dirname)):
        dirname = os.path.join(__models_dir__, dirname)
    weights = get_last_checkpoint(dirname) if weights is None else os.path.join(dirname, 'weights_%s.tar' % weights)
    config = toml.load(os.path.join(dirname, 'config.toml'))
    config = set_config_defaults(config, chunksize, batchsize, overlap, quantize)
    config["model"]["package"] = "relseg.transformer" # this is the change for lrp, because the location is usaully in the .toml file, changing this would break everything
    return _load_model(weights, config, device, half, use_koi)



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

from relseg import main_lrp_entry
__version__ = "0.1.0"

def main():
    # Create the main argument parser
    parser = main_lrp_entry.argparser()

    # Parse arguments and pass them to lrp.main()
    args = parser.parse_args()
    main_lrp_entry.main(args)

if __name__ == "__main__":
    main()


from bonito.io import *

def write_segmentation(read_id, sequence, segments, filename):
    assert len(sequence) == segments.shape[0], f"problem (read_id: {read_id}), number of bases and number of segments doesn't match: (bases: {len(sequence)}, segments:{segments.shape[0]})"
    with open(filename, "a") as f:
        for base, segment in zip(sequence, segments):
            f.write(f"{read_id}\t{base}")
            f.write(f"\t{int(segment[0])}\t{int(segment[1])}")
            f.write("\n")

def get_segment_filename():
    """
    Return the filename to use for the segment tsv.
    """
    stdout = realpath('/dev/fd/1')
    if sys.stdout.isatty() or stdout.startswith('/proc'):
        return 'segments.tsv'
    return '%s_segments.tsv' % splitext(stdout)[0]  

class LRP_Writer(Writer):
    def run(self):
        with CSVLogger(summary_file(), sep='\t') as summary:

            segment_filename = get_segment_filename()
            with open(segment_filename, "w") as f:
                f.write(f"read_id\tbase\tstart\tend\n")

            for read, res in self.iterator:

                seq = res['sequence']
                qstring = res.get('qstring', '*')
                mean_qscore = res.get('mean_qscore', mean_qscore_from_qstring(qstring))
                mapping = res.get('mapping', False)
                mods_tags = res.get('mods', [])

                samples = len(read.signal)
                read_id = read.read_id

                self.log.append((read_id, samples))

                if mean_qscore < self.min_qscore:
                    continue

                tags = [
                    f'RG:Z:{read.run_id}_{self.group_key}',
                    f'qs:i:{round(mean_qscore)}',
                    f'ns:i:{read.num_samples}',
                    f'ts:i:{read.trimmed_samples}',
                    *read.tagdata(),
                    *mods_tags,
                ]
                tags.append(f'mv:B:c,{encode_moves(res["moves"], res["stride"])}')

                if len(seq):
                    segments = res["segments"]
                    write_segmentation(read_id, seq, segments, segment_filename)

                    write_fastq(read_id, seq[::-1], qstring[::-1], fd=self.fd, tags=tags) # reverse to have real orientation

                    summary.append(summary_row(read, len(seq), mean_qscore, alignment=mapping))
                else:
                    logger.warn("> skipping empty sequence %s", read_id)



import logging
import types
from functools import lru_cache

logger = logging.getLogger(__name__)

import torch
import torch.nn.functional as F
from torch import nn
import math

# from bonito.lrp_folder.LRP_composites import ProjSwigluMultiplication, AttentionValueMatmul
from .RMSNorm import RMSNorm
from .RoPE import RotaryEmbedding

import relseg.lxt_adjustments as lf



from bonito.nn import from_dict, register, LinearCRFEncoder, MakeContiguous, Module, Permute, Serial



def deepnorm_params(depth):
    """
    Returns the DeepNorm (https://arxiv.org/abs/2203.00555) alpha and beta parameters.
    """
    alpha = round((2*depth)**0.25, 7)
    beta = round((8*depth)**(-1/4), 7)
    return alpha, beta


@lru_cache(maxsize=2)
def sliding_window_mask(seq_len, window, device):
    band = torch.full((seq_len, seq_len), fill_value=1.0)
    band = torch.triu(band, diagonal=-window[0])
    band = band * torch.tril(band, diagonal=window[1])
    band = band.to(torch.bool).to(device)
    return band



class MultiHeadAttention(Module):
    def __init__(self, d_model, nhead, qkv_bias=False, out_bias=True, rotary_dim=None, attn_window=None):
        super().__init__()
        assert d_model % nhead == 0, "d_model must be divisible by nhead"

        self.d_model = d_model
        self.nhead = nhead
        self.head_dim = d_model // nhead
        self.rotary_dim = self.head_dim if rotary_dim is None else rotary_dim
        
        self.Wqkv = torch.nn.Linear(d_model, 3 * d_model, bias=qkv_bias)

        self.out_proj = nn.Linear(d_model, d_model, bias=out_bias)

        self.rotary_emb_flash = RotaryEmbedding(self.rotary_dim, interleaved=False)

        self.attn_window = (-1, -1) if attn_window is None else tuple(attn_window)
        # self.matmul = AttentionValueMatmul()
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x):
        N, T, _ = x.shape

        qkv = self.Wqkv(x).view(N, T, 3, self.nhead, self.head_dim)
        v_val = qkv[:,:,2,:,:]
        q_val = qkv[:,:,0,:,:]
        k_val = qkv[:,:,1,:,:]

        q_val = self.rotary_emb_flash.apply_rotary_embedding_not_flash_x(q_val)
        k_val = self.rotary_emb_flash.apply_rotary_embedding_not_flash_x(k_val)


        attn_mask = sliding_window_mask(q_val.shape[1], self.attn_window, q_val.device)

        query, key, value = q_val.permute(0,2,1,3)[:,None,:,:,:], k_val.permute(0,2,1,3)[:,None,:,:,:], v_val.permute(0,2,1,3)[:,None,:,:,:]

        L, S = query.size(-2), key.size(-2)
        scale_factor = 1 / math.sqrt(query.size(-1))
        attn_bias = torch.zeros(L, S, dtype=query.dtype, device=query.device)
  

        if attn_mask is not None:
            attn_bias.masked_fill_(attn_mask.logical_not(), float("-Inf"))

        attn_weight = lf.matmul(query, key.transpose(-2, -1))
        attn_weight = lf.mul2(attn_weight, torch.tensor(scale_factor).detach())
        attn_weight = lf.add2(attn_weight, attn_bias.detach()) # IMPORTANT TODO here the relevance is lost because attn_bias is -Inf
        attn_weight = self.softmax(attn_weight)
        #attn_weight = torch.dropout(attn_weight, dropout_p, train=True) # LXT no dropout (dont need since were not training)
        
        # attn_out = self.matmul(attn_weight, value)
        attn_out = lf.matmul(attn_weight, value)
        attn_out = attn_out.permute(0, 1, 3, 2, 4)

        attn_out = attn_out.reshape(N, T, self.d_model)

        out = self.out_proj(attn_out)

        return out
    


class GatedMlp(nn.Module): # IMPORTANT simple implementation of mlp, should not be a problem for lxt, might need to think about activation functions
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        activation="sigmoid",
        bias1=True,
        bias2=True,
        multiple_of=128,
        return_residual=False,
        device=None,
        dtype=None,
    ):
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.activation = activation
        out_features = out_features if out_features is not None else in_features
        hidden_features = (
            hidden_features if hidden_features is not None else int(8 * in_features / 3)
        )
        hidden_features = (hidden_features + multiple_of - 1) // multiple_of * multiple_of
        self.return_residual = return_residual
        self.fc1 = nn.Linear(in_features, 2 * hidden_features, bias=bias1, **factory_kwargs)
        self.activation = activation
        self.fc2 = nn.Linear(hidden_features, out_features, bias=bias2, **factory_kwargs)
        self.silu = nn.SiLU(inplace=False)
        # self.swiglu_mul = ProjSwigluMultiplication()


    def forward(self, x): #IMPORTANT maybe call apply of swiglu?
        y = self.fc1(x)

        #swiglu:
        y, gate = y.chunk(2, dim=-1)
        # y = self.swiglu_mul(y, self.silu(gate))
        y = lf.mul2(y, self.silu(gate))
        y = self.fc2(y)
        return y if not self.return_residual else (y, x)
    
    


@register
class TransformerEncoderLayer(Module):
    def __init__(self, d_model, nhead, dim_feedforward, deepnorm_alpha, deepnorm_beta, attn_window=None):
        super().__init__()
        self.kwargs = {
            "d_model": d_model,
            "nhead": nhead,
            "dim_feedforward": dim_feedforward,
            "deepnorm_alpha": deepnorm_alpha,
            "deepnorm_beta": deepnorm_beta,
            "attn_window": attn_window
        }

        self.self_attn = MultiHeadAttention(
            d_model=d_model,
            nhead=nhead,
            qkv_bias=False,
            out_bias=True,
            attn_window=attn_window
        )
        self.ff = GatedMlp(
            d_model,
            hidden_features=dim_feedforward,
            activation="swiglu",
            bias1=False,
            bias2=False,
            multiple_of=1,
        )

        self.norm1 = RMSNorm(d_model)
        self.norm2 = RMSNorm(d_model)

        self.register_buffer("deepnorm_alpha", torch.tensor(deepnorm_alpha))
        self.reset_parameters()

    def reset_parameters(self):
        db = self.kwargs["deepnorm_beta"]
        d_model = self.kwargs["d_model"]
        torch.nn.init.xavier_normal_(self.ff.fc1.weight, gain=db)
        torch.nn.init.xavier_normal_(self.ff.fc2.weight, gain=db)
        torch.nn.init.xavier_normal_(self.self_attn.out_proj.weight, gain=db)
        torch.nn.init.xavier_normal_(self.self_attn.Wqkv.weight[2*d_model:], gain=db)
        torch.nn.init.xavier_normal_(self.self_attn.Wqkv.weight[:2*d_model], gain=1)

    def forward(self, x):
        x1 = self.self_attn(x)
        residuals_1 = lf.mul2(self.deepnorm_alpha.detach(), x)
        x1 = lf.add2(x1, residuals_1)
        x2 = self.norm1(x1) #IMPORTANT the xs have to be like this, also mind deepnorm_alpha*x

        x3 = self.ff(x2)
        residuals_2 = lf.mul2(self.deepnorm_alpha.detach(), x2)
        x3 = lf.add2(x3, residuals_2)
        x4 = self.norm2(x3)

        return x4

    def to_dict(self, include_weights=False):
        if include_weights:
            raise NotImplementedError
        return self.kwargs


def use_koi(self, **kwargs):
    # koi needs modified LinearCRFLayer settings
    def _expand_blanks(m):
        if isinstance(m, LinearCRFEncoder):
            m.expand_blanks = False
    self.encoder.apply(_expand_blanks)
    self.encoder = Serial([
        self.encoder,
        Permute([1, 0, 2]),
        MakeContiguous(),
    ])


def Model(config):
    model_config = {k: v for k, v in config["model"].items() if k != "package"}
    model = from_dict(model_config)
    model.config = config
    model.use_koi = types.MethodType(use_koi, model)
    return model


import matplotlib.pyplot as plt
import numpy as np
import torch

def plot_relevances(relevances, raw_signal, out_name):
    raw_signal = raw_signal.detach().cpu().numpy()
    relevances = relevances.detach().cpu().numpy()

    relevances = relevances.T
    fig, axs = plt.subplots((2))
    fig.set_size_inches(10,5)

    xmin = 0
    xmax = 1000
    old_segment = 0
    old_relevance = 0
    counter = 0
    axs[0].plot(raw_signal, color="black", linewidth=0.3, label="raw_data")
    for i,relevance in enumerate(relevances):
        # relevance = np.abs(relevance)
        # relevance = relevance/np.max(relevance)
        # if sum(relevance) < 500:

        axs[1].plot(relevance, label="relevance", alpha=0.7)

        # if np.argmax(relevance)+20<old_segment:
        #     if counter==3:
        #         axs[1].plot(relevance, label="relevance", alpha=0.7, color="orange")
        #         axs[0].vlines(np.argmax(relevance), -3, 3, color="orange")

        #         axs[1].plot(old_relevance, label="relevance", alpha=0.7, color="blue")
        #         axs[0].vlines(old_segment, -3, 3, color="blue")

        #         axs[1].plot(old_old_relevance, label="relevance", alpha=0.7, color="green")
        #         axs[0].vlines(old_old_segment, -3, 3, color="green")       

        #     counter +=1

        # old_old_relevance = old_relevance
        # old_old_segment = old_segment
        # old_segment = np.argmax(relevance)
        # old_relevance = relevance

        # print(relevance)

    # axs[1].plot(np.abs(relevances[40])/np.max(np.abs(relevances[40])))

    axs[0].set_title("Signal")
    axs[0].set_ylabel("Normed Current")
    axs[1].set_ylabel("Relevance")
    axs[1].set_title("Relevance")
    axs[1].set_xlabel("Data Points")

    axs[1].set_xlim(xmin,xmax)
    axs[0].set_xlim(xmin,xmax)
    # axs[1].set_ylim(None, 3)
    fig.tight_layout()

    plt.savefig(out_name, dpi=200, format="pdf")

if __name__ == "__main__":
    relevances = torch.load("relevance/relevances_0.pkl")
    signal = torch.load("relevance/signals_0.pkl")

    plot_relevances(relevances[0,:,:], signal[0,0,:], "plots/test.pdf")

    # plot_relevances(relevances[1,:,14:15], signal[1,0,:], "plots/lrp_single.pdf")
    # plot_relevances(relevances[1,:,:], signal[1,0,:], "plots/lrp_all.pdf")

    # plot_relevances(relevances[1,:,:], signal[1,0,:], "plots/lrp_problem.pdf")
    # plot_relevances(relevances[1,:,:], signal[1,0,:], "plots/lrp_wrong_order.pdf") # counter == 3


import os
import re
from setuptools import setup, find_packages


__pkg_name__ = 'relseg'
require_file = 'requirements.txt'
package_name = __pkg_name__

verstrline = open(os.path.join(__pkg_name__, '__init__.py'), 'r').read()
vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(vsre, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in "{}/__init__.py".'.format(__pkg_name__))


with open(require_file) as f:
    requirements = [r.split()[0] for r in f.read().splitlines()]

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


setup(
    name=package_name,
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='LKoehler',
    python_requires='==3.10',
    entry_points={
        'console_scripts': [
            '{0} = {0}:main'.format(__pkg_name__)
        ]
    },
)



import lxt.functional as lf
import torch
import torch.fx
from torch.autograd import Function
import torch.nn.functional as F

EPSILON = 1e-6

def add2(input_a, input_b):
    return add2_tensors_fn.apply(input_a, input_b, False, EPSILON)

def matmul(input_a, input_b):
    return lf.matmul(input_a, input_b, epsilon=EPSILON)

mul2 = lf.mul2


class add2_tensors_fn(Function):
    """
    Standard Epsilon-LRP rule for elementwise addition (along all dimensions) of two tensors according to the Equation 8 of the paper
    'AttnLRP: Attention-Aware Layer-wise Relevance Propagation for Transformers'

    Parameters:
    -----------
    input_a: torch.Tensor
        The first input tensor
    input_b: torch.Tensor
        The second input tensor
    inplace: bool
        Whether to perform the operation in place during the backward pass, will overwrite the relevance at the output
    epsilon: float
        Small value to stabilize the denominator
    """
    
    @staticmethod
    def forward(ctx, input_a, input_b, inplace=False, epsilon=1e-6):
    
        outputs = input_a + input_b
        if any([inp.requires_grad for inp in (input_a, input_b)]):
            ctx.save_for_backward(input_a, input_b)
            ctx.epsilon, ctx.inplace = epsilon, inplace

        return outputs

    @staticmethod
    @lf.conservation_check_wrap
    def backward(ctx, *out_relevance):

        #TODO: replace for conservation check with requires grad stuff

        input_a, input_b = ctx.saved_tensors


        if ctx.inplace:
            relevance_norm = out_relevance[0].div_(lf._stabilize(input_a + input_b, epsilon=ctx.epsilon, inplace=True))

            relevance_a = relevance_norm * input_a
            relevance_b = relevance_norm.mul_(input_b)

        else:
            relevance_norm = out_relevance[0] / lf._stabilize(input_a + input_b, epsilon=ctx.epsilon, inplace=True)

            relevance_a = relevance_norm * input_a
            relevance_b = relevance_norm * input_b

        relevance_a = torch.where(torch.isnan(relevance_a), torch.tensor(0).to(relevance_a), relevance_a) #stabilize attention mask, this is what is changed from orig. lxt
        relevance_b = torch.where(torch.isnan(relevance_b), torch.tensor(0).to(relevance_b), relevance_b) 

        return (relevance_a, relevance_b, None, None)


import torch
import numpy as np
from bonito.util import chunk, batchify, unbatchify, half_supported
from koi.decode import to_str
from scipy.signal import find_peaks
from relseg.stitching import stitch_results
from relseg.ctc import run_beam_search, run_viterbi
from relseg.lxt_composites import lxt_comp, zennit_comp1,zennit_comp2, zennit_comp3, zennit_comp4, zennit_comp_first_conv

def fmt(stride, attrs, trimmed_samples, rna=False):
    segments = attrs['segments']
    segments[segments[:,1] !=-1, :] += trimmed_samples
    return {
        'stride': stride,
        'moves': attrs['moves'].numpy(),
        'segments': segments,
        'qstring': to_str(attrs['qstring']), # IMPORTANT dont flip sequence etc like in basecaller because otherwirse they are not mappable to the raw signal
        'sequence': to_str(attrs['sequence']), #IMPORTANT here int is translated to ascii (ACGT)
    }




def register(model, dummy_input, input_name): # the imported composites are used
    model.eval()

    parent = lxt_comp.register(model)#, dummy_inputs={input_name: dummy_input}, verbose=False) # "input" / "x" TESTVALUE



    if "namedserial" == model[0].name: # using beamsearch adds a namedserial ontop of conv
        zennit_comp_first_conv.register(model[0][0][0])
        zennit_comp1.register(model[0][0][1])
        zennit_comp2.register(model[0][0][2])
        zennit_comp3.register(model[0][0][3])
        zennit_comp4.register(model[0][0][4])
    else:
        zennit_comp_first_conv.register(model[0][0])
        zennit_comp1.register(model[0][1])
        zennit_comp2.register(model[0][2])
        zennit_comp3.register(model[0][3])
        zennit_comp4.register(model[0][4])

    return model

def batch_positions(positions):
    """
    get move positions (index where a base is called) and create a tensor with those indices
    then backward() with one base for each sample of the batch instead of each individually
    if one sample of the batch has fewer bases/moves than the other then its filled with -1
    """
    result = torch.zeros_like(positions, dtype=torch.long)-1
    most_moves_in_batch = 0
    for i, sample in enumerate(positions):
        moves = torch.argwhere(sample==1).squeeze()
        l = len(moves)
        result[i, :l] = moves
        if l > most_moves_in_batch:
            most_moves_in_batch = l

    result = result[:,:most_moves_in_batch]
    return result



def batched_lrp_loop(data, y, batched_positions, traceback=None):
    full_batch_indices = torch.arange(data.shape[0], dtype=torch.long)

    for positions in batched_positions.T:
        batch_indices_filtered = full_batch_indices[positions!=-1] # batch_indices_filtered has the actually viable positions of bases for later
        # if there are no more bases in a sample of a batch, then it just computes the last one over and over, because it cant do backward() without a "full" batch
        if traceback != None:
            kmer = traceback[full_batch_indices, positions]
            y_current = y[full_batch_indices, positions, kmer].mean()*1000 # TESTVALUE so gradient doesnt go to 0 at end

        else:
            y_current = y[full_batch_indices, positions, :].mean()*1000

        data.grad = None
        y_current.backward(retain_graph=True)
        relevance = data.grad[batch_indices_filtered, 0,:]

        yield (relevance, batch_indices_filtered, positions[positions!=-1])

RELEVANCE_INDEX = 0  # Global counter
def wrapped_batched_lrp_loop(data, y, batched_positions, traceback=None, save_relevance=False):
    global RELEVANCE_INDEX

    if save_relevance:
        batchsize, _, seq_len = data.shape
        relevances = torch.zeros((batchsize,seq_len, batched_positions.shape[1]), dtype=data.dtype, device=data.device)
    for i, (relevance, batch_indices_filtered, positions) in enumerate(batched_lrp_loop(data, y, batched_positions, traceback)):
        if save_relevance:
            relevances[batch_indices_filtered, :, i] = relevance
        yield (relevance, batch_indices_filtered, positions)
    if save_relevance: # not very elegant way of saving the relevance
        torch.save(relevances, f"relevance/relevances_{RELEVANCE_INDEX}.pkl")
        torch.save(data, f"relevance/signals_{RELEVANCE_INDEX}.pkl")
        RELEVANCE_INDEX += 1
        # assert False
    # TODO rethink this, how do i appropriately save it, now it keeps getting overwritten



def segmentation_loop(relevance_gen, segmentation_function_out_shape, batchsize, downsampled_size):
    segmentation_function, segment_shape = segmentation_function_out_shape
    segments_batch = torch.zeros((batchsize, downsampled_size, segment_shape[0], segment_shape[1]), dtype=torch.float) - 2 # this tensor is longer than needed and will never be filled because n_moves is shorter than downsampled_size, but its easier to unbatchify

    for i, (relevance, batch_indices, motif_indices) in enumerate(relevance_gen):

        segment_indices = segmentation_function(relevance).to("cpu")
        z = torch.zeros((batchsize, segment_shape[0], segment_shape[1]), dtype=torch.float) - 2
        z[batch_indices] = segment_indices # if a sample in the batch no longer has moves then just keep adding 0 as segment until all samples in batch have no more moves
        segments_batch[batch_indices, motif_indices,:,:] = segment_indices

    return segments_batch

def peak_segmentation(number_peaks):
    segment_shape = (number_peaks, 2)

    def func(relevances):
        result = torch.zeros((relevances.shape[0], number_peaks, 2)) - 1 # not -2 because i dont want to delete later

        for i,relevance in enumerate(relevances):

            relevance = torch.nn.functional.pad(relevance.abs(), (1,1), "constant", 0)  # pad at beginning and end if peak is at position 0 then it will not find the peak because there is nothing to the left
            relevance = relevance / (relevance.max())
            peaks = find_peaks(relevance.detach().cpu(), distance=3, height=0.2)
            peaks = np.array([peaks[0]-1, peaks[1]["peak_heights"]]) # -1 because of the padding

            peaks = peaks[:,np.argsort(peaks, 1)[1]]
            peaks = np.flip(peaks, axis=1).T
            peaks = torch.from_numpy(peaks.copy())
            peaks = peaks[:number_peaks].to(torch.float)
            result[i, :peaks.shape[0], torch.arange(2)] = peaks

            
        return result
    
    return func, segment_shape


def argmax_segmentation():
    def func(relevances):
        indices = torch.argmax(torch.abs(relevances), dim=1, keepdim=False)
        return indices.view(-1,1,1).to(torch.float)
    return func, (1,1)



def forward_and_lrp(model_enc, input_signal, seqdist, search_algorithm,  save_relevance=False): #MAIN LRP FUNCTION
    device = next(model_enc.parameters()).device
    dtype = torch.float16 if half_supported() else torch.float32
    input_signal = input_signal.to(dtype).to(device)

    scores = model_enc(input_signal.requires_grad_(True))

    if search_algorithm == "viterbi":
        search_result = run_viterbi(scores, seqdist, len(seqdist.alphabet), seqdist.n_base)
        scores = scores.permute(1,0,2)
        kmers = search_result["traceback"]
    else:
        search_result = run_beam_search(scores) # chageable
        kmers = None
        
    batchsize, downsampled_len, _ = scores.shape
    moves = search_result["moves"]
    batched_moves = batch_positions(moves) # -> shape [#moves, nbatches]


    relevance_gen = wrapped_batched_lrp_loop(input_signal, scores, batched_moves, kmers, save_relevance)

    segments = segmentation_loop(relevance_gen, peak_segmentation(5), batchsize, downsampled_len)
    search_result["segments"] = segments
    return search_result





def basecall_and_lrp(model, reads, search_algorithm, chunksize=4000, overlap=100, batchsize=4,
             reverse=False, rna=False, save_test_relevance=False):
    """
    Basecalls a set of reads.
    """
    # reads = (read for read in reads if read.read_id == "3b73c140-6ce2-461a-a36a-af28e2afb77f")

    chunks = (
        ((read, 0, read.signal.shape[-1]), chunk(torch.from_numpy(read.signal), chunksize, overlap))
        for read in reads
    )

    batches = (batchify(chunks, batchsize=batchsize))

    dummy_input = torch.randn((batchsize,1,chunksize), device=next(model.parameters()).device)
    input_name = "input" if search_algorithm == "viterbi" else "x"
    model_enc = register(model.encoder, dummy_input, input_name)

    scores = (
        (read, forward_and_lrp(model_enc, batch, model.seqdist, search_algorithm, save_test_relevance)) for read, batch in batches
    )

    results = (
        (read, stitch_results(scores, end - start, chunksize, overlap, model.stride, "", reverse))
        for ((read, start, end), scores) in unbatchify(scores)
    )

    return (
        (read, fmt(model.stride, attrs, read.trimmed_samples, rna))
        for read, attrs in results
    )


